# SPDX-FileCopyrightText: Copyright 2021, Siavash Ameli <sameli@berkeley.edu>
# SPDX-License-Identifier: BSD-3-Clause
# SPDX-FileType: SOURCE
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the license found in the LICENSE.txt file in the root
# directory of this source tree.


"""
Some of the test files produce mstplotlib figures. During the tests (especially
on a server), non of the figures should be plotted in interactive mode. Rather,
the figures should be saved into ``SVG`` file. To do so, the utility file
``/ortho/Utilities/PlotUtilities.py`` handles this as follows:

* If the environment variable ``DISPLAY`` does not exist or is ``''``, or
* If the environment variable ``ortho_NO_DISPLAY`` is set to ``'True'``,

then, the ``matplotlib`` is imported with ``'agg'`` mode, which prevents the
interactive plot, and instead, the plots are saved as ``SVG`` files.

Currently, the following test files produce figures:

    * ``test_correlation_matrix.py``
    * ``test_interpolate_traceinv.py``.

In the beginning of the above test files, and before importing ``ortho``,
these lines should be added:

::

    import os
    os.environ['ortho_NO_DISPLAY'] = 'True'

However, the above approach yet does not prevent the interactive plotting when
*all* test files are tested using ``pytest``. When the command

::

    pytest

is used, it runs all test files, including those that do not produce figure.
Because those test files do not set the environment variable
``ortho_NO_DISPLAY``, the ``matplotlib`` in the module ``plot_utilities.py``
will be loaded without the ``agg`` mode, which then affects the other test
files that define such environment variable, since the ``matplotlib`` is
already loaded.

To prevent this issue, either the above environment variable should be loaded
on *all* test files (regardless they produce figures or not), or to define it
here in this ``__init__.py`` file.
"""

# For plotting matrix, we disable interactive display
import os
os.environ['ORTHO_NO_DISPLAY'] = 'True'   # define before importing ortho



# SPDX-FileCopyrightText: Copyright 2021, Siavash Ameli <sameli@berkeley.edu>
# SPDX-License-Identifier: BSD-3-Clause
# SPDX-FileType: SOURCE
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the license found in the LICENSE.txt file in the root directory
# of this source tree.


# =======
# Imports
# =======

from .orthogonalization_utilities import gram_schmidt_process
from .orthogonalization_utilities import print_coeffs_of_functions
from .orthogonalization_utilities import check_mutual_orthonormality
from .orthogonalization_utilities import get_symbolic_coeffs
from .orthogonalization_utilities import get_numeric_coeffs
from .plot_utilities import plot_functions


# ====================
# Orthogonal Functions
# ====================

class OrthogonalFunctions(object):
    """
    Generates a set of orthonormal functions.

    Parameters
    ----------

        start_index : int, default=1
            The index of the starting function, :math:`i_0`.

        num_func : int, default=9
            Number of orthogonal functions to generate

        end_interval : float, default=1
            The right interval of orthogonality, :math:`L`.

        vebose : boolean, default=False
            Prints the generated functions. An example of output is

            ::

                phi_1(t) =  sqrt(x)
                phi_2(t) =  sqrt(6)*(5*x**(1/3) - 6*sqrt(x))/3
                phi_3(t) =  sqrt(2)*(21*x**(1/4) - 40*x**(1/3) + 20*sqrt(x))/2

    Attributes
    ----------

        sym_phi : sympy obj
            Orthogonal functions :math:`\\phi_i^{\\perp}`

        sym_alpha : sympy obj
            Coefficients :math:`\\alpha_i`

        sym_coeffs : sympy obj
            Coefficients :math:`a_{i,j}`

        alpha : list
            Coefficients :math:`\\alpha_i`

        coeffs : list of lists
            Coefficients :math:`a_{i,j}`

    Notes
    -----

    The orthogonal functions, :math:`\\phi_i^{\\perp}`, are generated based on
    the set of non-orthonormal functions :math:`\\phi_i` defined by the
    inverse-monomials

    .. math::

        \\phi_i(t) = t^{\\frac{1}{i+1}}, \\qquad i = i_0,\\dots,i_0+n.

    The orthonormalized functions :math:`\\phi_i^{\\perp}` are the linear
    combination of the functions :math:`\\phi_i` by

    .. math::

        \\phi_i^{\\perp}(t) = \\alpha_i \\sum_{j = i_0}^{i_0+n} a_{ij}
        \\phi_j(t), \\qquad i = i_0,\\dots,i_0+n.

    The functions :math:`\\phi_i^{\\perp}` are orthonormal in the interval
    :math:`t \\in [0,L]` with respect to the weight function
    :math:`w(t) = t^{-1}`. That is,

    .. math::

        \\langle \\phi_i^{\\perp},\\phi_j^{\\perp} \\rangle_{L^2([0,L],
        \\mathrm{d}t/t)} = \\int_0^L \\phi_i^{\\perp}(t) \\phi_j^{\\perp}(t)
        \\frac{\\mathrm{d}t}{t} = \\delta_{ij},

    where :math:`\\delta_{ij}` is the Kronecker delta function. The orthogonal
    functions are generated by `Gram-Schmidt orthogonalization process
    <https://en.wikipedia.org/wiki/Gram%E2%80%93Schmidt_process>`__.

    Methods
    -------
    check
    print
    plot
    """

    # ----
    # Init
    # ----

    def __init__(
            self,
            start_index=1,
            num_func=9,
            end_interval=1,
            verbose=False):
        """
        Parses the user inputs and sets the member data of the object.
        """

        # Number of functions
        self.num_func = num_func
        self.start_index = start_index
        self.end_interval = end_interval
        self.verbose = verbose

        # Check arguments
        if self.num_func < 1:
            print('The option "num_func" should be at least 1.')
            exit(1)
        if self.start_index < 0:
            print('"start_index" should be at least 1.')
            exit(1)
        if end_interval <= 0.0:
            print('"end_interval" should be greater than zero.')
            exit(1)

        # interval
        self._interval = [0, end_interval]

        # Generate list of functions
        self._phi_orthonormalized_list = self._process()
        self.sym_phi = self._phi_orthonormalized_list

        # Get coeffs of symbolic functions
        self.sym_alpha, self.sym_coeffs = get_symbolic_coeffs(
                self._phi_orthonormalized_list, self.start_index)

        # Get numeric values of coefficients
        self.alpha, self.coeffs = get_numeric_coeffs(
                self.sym_alpha, self.sym_coeffs)

    # -------
    # Process
    # -------

    def _process(self, verbose=False):
        """
        Computes the set of orthonormalized functions.
        """

        # Generate a list of symbolic orthonormal functions
        phi_orthonormalized_list = gram_schmidt_process(
                self.num_func,
                self.start_index,
                self._interval,
                verbose=verbose)

        return phi_orthonormalized_list

    # -----
    # Check
    # -----

    def check(self, verbose=False):
        """
        Check the mutual orthogonality of the functions.

        Parameters
        ----------

            verbose : boolean, default=False
                If `True`, prints the matrix of mutual inner product of
                functions. An example of the output is

                ::

                    [[1 0 0 0 0 0 0 0 0]
                     [0 1 0 0 0 0 0 0 0]
                     [0 0 1 0 0 0 0 0 0]
                     [0 0 0 1 0 0 0 0 0]
                     [0 0 0 0 1 0 0 0 0]
                     [0 0 0 0 0 1 0 0 0]
                     [0 0 0 0 0 0 1 0 0]
                     [0 0 0 0 0 0 0 1 0]
                     [0 0 0 0 0 0 0 0 1]]

        Returns
        -------

            status : boolean
                If `True`, the generated functions are all mutually
                orthonormal. Otherwise returns `False`.
        """

        if self._phi_orthonormalized_list is None:
            raise RuntimeError('Call Process() first.')

        # Check orthonormality of functions
        status = check_mutual_orthonormality(
                self._phi_orthonormalized_list,
                self._interval,
                verbose=verbose)

        return status

    # -----
    # Print
    # -----

    def print(self):
        """
        Print coeffs of Functions.

        An Example of output is

        ::

              i      alpha_i                                    a_[ij]
            ------  ----------   ---------------------------------------------
            i = 1:  +sqrt(2/2)   [1                                          ]
            i = 2:  -sqrt(2/3)   [6,   -5                                    ]
            i = 3:  +sqrt(2/4)   [20,  -40,    21                            ]
            i = 4:  -sqrt(2/5)   [50,  -175,   210,   -84                    ]
            i = 5:  +sqrt(2/6)   [105, -560,   1134,  -1008,   330           ]
            i = 6:  -sqrt(2/7)   [196, -1470,  4410,  -6468,   4620,   -1287 ]
        """

        if self._phi_orthonormalized_list is None:
            raise RuntimeError('Call Process() first.')

        print_coeffs_of_functions(
                self.sym_coeffs,
                self.start_index)

    # ----
    # Plot
    # ----

    def plot(self):
        """
        Plot the generated functions.
        """

        if self._phi_orthonormalized_list is None:
            raise RuntimeError('Call Process() first.')

        # Plot Functions
        plot_functions(
                self._phi_orthonormalized_list,
                self.start_index,
                self._interval)



# SPDX-FileCopyrightText: Copyright 2021, Siavash Ameli <sameli@berkeley.edu>
# SPDX-License-Identifier: BSD-3-Clause
# SPDX-FileType: SOURCE
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the license found in the LICENSE.txt file in the root directory
# of this source tree.


# =======
# Imports
# =======

import os
import sympy
import numpy
from distutils.spawn import find_executable
import matplotlib
import matplotlib.pyplot as plt
from .declarations import t


# =============
# Plot Settings
# =============

def plot_settings():
    """
    General settings for the plot.
    """

    # Color palette
    import seaborn as sns
    # sns.set()

    # Axes font size
    sns.set(font_scale=1.2)

    # LaTeX
    if find_executable('latex'):
        plt.rc('text', usetex=True)

    # Style sheet
    sns.set_style("white")
    sns.set_style("ticks")

    # Font (Note: this should be AFTER the plt.style.use)
    plt.rc('font', family='serif')
    plt.rcParams['svg.fonttype'] = 'none'  # text in svg file is text not path.


# ==============
# Plot Functions
# ==============

def plot_functions(phi_orthonormalized_list, start_index, interval):
    """
    Plots the generated functions, also saves the plots as both ``svg`` and
    ``pdf`` format.

    :param phi_orthonormalized_list: The list of generated functions. Each
        entry is a ``sympy`` object.
    :param: list

    :param start_index: The indet of the first function.
    :type start_index: int

    :param Interval: The right side of the interval of the domain of the
        functions.
    :param Interval: float
    """

    # Run plot settings
    plot_settings()

    # Axis
    t_array = numpy.logspace(-7, numpy.log10(interval[1]), 1000)

    # Evaluate functions
    num_functions = len(phi_orthonormalized_list)

    f = numpy.zeros((num_functions, t_array.size), dtype=float)
    for j in range(num_functions):
        f_lambdify = sympy.lambdify(t, phi_orthonormalized_list[j], 'numpy')
        f[j, :] = f_lambdify(t_array)

    # Plot
    fig, ax = plt.subplots(figsize=(7, 4.8))
    for j in range(num_functions):
        ax.semilogx(t_array, f[j, :],
                    label=r'$i = %d$' % (j+start_index))

    ax.legend(ncol=3, loc='lower left', borderpad=0.5, frameon=False)
    ax.set_xlim([t_array[0], t_array[-1]])
    ax.set_ylim([-1, 1])
    ax.set_yticks([-1, 0, 1])
    ax.set_xlabel(r'$t$')
    ax.set_ylabel(r'$\phi_i^{\perp}(t)$')
    ax.set_title('Orthogonalized inverse-monomial functions')
    ax.grid(axis='y')

    # Get the root directory of the package (parent directory of this script)
    file_dir = os.path.dirname(os.path.realpath(__file__))
    parent_dir = os.path.dirname(file_dir)
    second_parent_dir = os.path.dirname(parent_dir)

    # Try to save in the docs/images directory. Check if exists and writable
    save_dir = os.path.join(second_parent_dir, 'docs', 'images')
    if (not os.path.isdir(save_dir)) or (not os.access(save_dir, os.W_OK)):

        # Write in the current working directory
        save_dir = os.getcwd()

    # Save plot in both svg and pdf format
    if os.access(save_dir, os.W_OK):
        save_fullename_svg = os.path.join(save_dir, 'orthogonal_functions.svg')
        save_fullename_pdf = os.path.join(save_dir, 'orthogonal_functions.pdf')
        plt.savefig(save_fullename_svg, transparent=True, bbox_inches='tight')
        plt.savefig(save_fullename_pdf, transparent=True, bbox_inches='tight')
        print('')
        print('Plot saved to "%s".' % (save_fullename_svg))
        print('Plot saved to "%s".' % (save_fullename_pdf))
    else:
        print('Cannot save plot to %s. Directory is not writable.' % save_dir)

    # If no display backend is enabled, do not plot in the interactive mode
    if matplotlib.get_backend() != 'agg':
        plt.show()



# SPDX-FileCopyrightText: Copyright 2021, Siavash Ameli <sameli@berkeley.edu>
# SPDX-License-Identifier: BSD-3-Clause
# SPDX-FileType: SOURCE
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the license found in the LICENSE.txt file in the root directory
# of this source tree.


# =======
# Imports
# =======

import sympy


# ==========================
# Declare symbolic variables
# ==========================

n = sympy.symbols('n', integer=True, positive=True)
t = sympy.symbols('t', real=True, positive=True)



# SPDX-FileCopyrightText: Copyright 2021, Siavash Ameli <sameli@berkeley.edu>
# SPDX-License-Identifier: BSD-3-Clause
# SPDX-FileType: SOURCE
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the license found in the LICENSE.txt file in the root directory
# of this source tree.


# =======
# Imports
# =======

import os
import sys
import getopt


# =============
# Print Version
# =============

def print_version():
    """
    Prints the version of the code given in.
    """

    # Get the parent directory of this script
    file_directory = os.path.dirname(os.path.realpath(__file__))
    parent_directory = os.path.dirname(file_directory)

    # Find the version
    version_dummy = {}
    exec(open(os.path.join(parent_directory, '__version__.py'), 'r').read(),
         version_dummy)
    version = version_dummy['__version__']
    del version_dummy

    # String to print
    version_string = \
        """
Version %s
        """ % (version)

    print(version_string)


# =============
# Print License
# =============

def print_license():
    """
    Prints the license and author info.
    """

    license_string = \
        """
Author:
Siavash Ameli
University of California, Berkeley
April 27, 2020

License: BSD-3-Clause
        """

    print(license_string)


# ===========
# Print Usage
# ===========

def print_usage(exec_name):
    """
    Prints the usage for the stand-alone application.
    """

    usage_string = \
        """
Usage:
$ %s  [options]
        """ % (exec_name)

    options_string = \
        """
Optional arguments:

-h --help                   Prints this help message.
-v --version                Prints version
-l --license                Prints author info, citation and license.
-n --num-func[=int]         Number of orthogonal functions to generate.
                            Positive integer. Default is 9.
-s --start-func[=int]       Starting function index. Non-negative integer.
                            Default is 1.
-e --end-interval[=float]   End of the interval of functions domains. Real
                            number greater than zero. Default is 1.
-c --check                  Checks orthogonality of generated functions.
-p --plot                   Plots generated functions, also saves the plot as
                            pdf and svg file in the current directory.
            """

    example_string = \
        """
Description:

This script generates a set of orthonormal functions, called phi_perp, based on
the set of non-orthonormal functions

    phi_i(t) = t**(1/(i+i)),      i = I, ... , I+N

The orthonormalized functions phi_perp_i are linear combination of the
functions phi_i, as

    phi_perp_i(t) = alpha_i * sum_{j=I}^{I+N} a_{ij} phi_j(t)

The functions phi_perp are orthonormal in the interval [0, L] with respect to
weight w(t) = 1/t. That is

    int_0^L phi_perp_i(t) phi_perp_j(t) 1/t dt = delta_{ij}

where delta_{ij} is the Kronecker delta function.

The script can be configured as follows:

   Variable     Variable in script              Option
   --------    --------------------    -------------------------
      I         "StartFunctionIndex"    "-s", or "--start-func"
      N         "NumFunctions"          "-n", or "--num-func"
      L         "EndInterval"           "-e", or "--end-interval"

Output:

1. Prints the symbolic functions
2. Prints a human readable coefficients, "alpha" and "a" of the functions
3. Prints a matrix of mutual inner product of functions to check orthogonality
   (with option "-c")
4. Plots the set of functions (with option "-p")
5. Saves the plot as pdf and svg in the current directory (with option "-p")

Examples:

1. Generate nine orthogonal functions from function index 1 to 9
    $ %s

2. Generate seven orthogonal functions from function index 1 to 8
    $ %s -n 8

3. Generate nine orthogonal functions from function index 0 to 8
    $ %s -s 0

4. Generate nine set of orthogonal functions starting from function 1, that are
   orthonormal in the interval [0, 10]
    $ %s -e 10

5. Check orthogonality of each two function, and plot the orthonormal functions
   and save the plot to pdf and svg
    $ %s -c -p

6. A complete example:
    $ %s -n 9 -s 1 -e 1 -c -p

        """ % (exec_name, exec_name, exec_name, exec_name, exec_name,
               exec_name)

    print(usage_string)
    print(options_string)
    print(example_string)


# ===============
# Parse Arguments
# ===============

def parse_arguments(argv, test=False):
    """
    Parses the argument of the executable.

    :param argv: list of keys and user inputs.
    :type: list

    :param Test: If ``True``, the program is not exited after options ``-h``,
        ``-v``, and ``-l``.
        If ``False``, the program is exited by the above options.
    :type Test: bool
    """

    # Initialize variables (defaults)
    arguments = {
        'check_orthogonality': False,
        'plot_flag': False,
    }

    try:
        opts, args = getopt.getopt(argv[1:], "hvln:s:e:cp", [
            "help", "version", "license", "num-func=", "start-func=",
            "end-interval=", "check", "plot"])
    except getopt.GetoptError:
        print('Invalid option entered.')
        print_usage(argv[0])
        sys.exit(2)

    # Assign options
    for opt, arg in opts:

        if opt in ('-h', '--help'):
            print_usage(argv[0])
            if not test:
                sys.exit(0)
        elif opt in ('-v', '--version'):
            print_version()
            if not test:
                sys.exit(0)
        elif opt in ("-l", "--license"):
            print_license()
            if not test:
                sys.exit(0)
        elif opt in ("-n", "--num-func"):
            arguments['num_func'] = int(arg)
        elif opt in ("-s", "--start-func"):
            arguments['start_index'] = int(arg)
        elif opt in ("-e", "--end-interval"):
            arguments['end_interval'] = float(arg)
        elif opt in ("-c", '--check'):
            arguments['check_orthogonality'] = True
        elif opt in ("-p", "--plot"):
            arguments['plot_flag'] = True

    return arguments



# SPDX-FileCopyrightText: Copyright 2022, Siavash Ameli <sameli@berkeley.edu>
# SPDX-License-Identifier: BSD-3-Clause
# SPDX-FileType: SOURCE
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the license found in the LICENSE.txt file in the root directory
# of this source tree.


# =======
# Imports
# =======

import os
import fnmatch

__all__ = ['recursive_glob']


# =======================
# split all parts of path
# =======================

def _split_all_parts_of_path(path):
    """
    Splits all parts of a path. For example, the path

        '../build/lib.linux-x86_64-3.8/Module/Submodule/lib.so'

    will be split to the list

        ['..','build','lib.linux-x86_64','Module','Submodule','lib.so']

    :param path: A file or directory path
    :type path: string

    :return: The list of strings of split path.
    :rtype: list(string)
    """

    all_parts = []

    # Recursion
    while True:

        # Split last part
        parts = os.path.split(path)

        if parts[0] == path:
            all_parts.insert(0, parts[0])
            break

        elif parts[1] == path:
            all_parts.insert(0, parts[1])
            break

        else:
            path = parts[0]
            all_parts.insert(0, parts[1])

    return all_parts


# =========================
# remove duplicates in list
# =========================

def _remove_duplicates_in_list(list):
    """
    Removes duplicate elements in a list.

    :param list: A list with possibly duplicate elements.
    :type list: list

    :return: A list which elements are not duplicate.
    :rtype: list
    """

    shrinked_list = []
    for element in list:
        if element not in shrinked_list:
            shrinked_list.append(element)

    return shrinked_list


# ==============
# recursive glob
# ==============

def recursive_glob(directory, patterns):
    """
    Recursively searches all subdirectories of a given directory and looks for
    a list of patterns. If in a subdirectory, one of the patterns is found, the
    name of the first immediate subdirectory (after Directory) is returned in a
    list.

    For example, is the pattern is '*.so',

    Directory
    |
    |-- SubDirectory-1
    |   |
    |   |--File-1.so
    |
    |-- SubDirectory-2
    |   |
    |   |--SubSubDirectory-2
    |      |
    |      |--File-2.so
    |
    |-- SubDirectory-3
    |   |
    |   |--File-3.a

    This code outputs ['SubDirectory-1','SubDirectory-2']. Note that the
    `SubSubDirectory-2` is not in the output, since it is part of the
    `SubDirectory-2'. That is, this code only outputs the first children
    subdirectory if within that subdirectory a match of pattern is found.

    .. note::

        In python 3, this function ``glob(dir,recursive=True)`` can be simply
        used. However, the recursive glob is not supproted in python 2 version
        of ``glob``, hence this function is written.

    :param directory: The path of a directory.
    :type directory: string

    :param patterns: A list of string as regex pattern, such as
        ['*.so','*.dylib','*.dll']
    :type patterns: list(string)

    :return: List of first-depth subdirectories that within them a match of
        pattern is found.
    :rtype: list(string)
    """

    # Find how many directory levels are in the input Directory path
    directory_depth = len(_split_all_parts_of_path(directory))

    subdirectories = []

    for root, dirname, filenames in os.walk(directory):
        for pattern in patterns:
            for filename in fnmatch.filter(filenames, pattern):
                subdirectories.append(
                        _split_all_parts_of_path(root)[directory_depth])

    return _remove_duplicates_in_list(subdirectories)



# SPDX-FileCopyrightText: Copyright 2021, Siavash Ameli <sameli@berkeley.edu>
# SPDX-License-Identifier: BSD-3-Clause
# SPDX-FileType: SOURCE
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the license found in the LICENSE.txt file in the root directory
# of this source tree.


from ._orthogonal_functions import OrthogonalFunctions

__all__ = ['OrthogonalFunctions']

from .__version__ import __version__                               # noqa: F401



#! /usr/bin/env python

# SPDX-FileCopyrightText: Copyright 2021, Siavash Ameli <sameli@berkeley.edu>
# SPDX-License-Identifier: BSD-3-Clause
# SPDX-FileType: SOURCE
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the license found in the LICENSE.txt file in the root directory
# of this source tree.


# =======
# Imports
# =======

from ortho._parse_arguments import parse_arguments


# ====================
# Test Parse arguments
# ====================

def test_parse_arguments():

    # Mock user argument
    argv1 = ['', '-n', '8', '-s', '1', '-e', '1', '-c', '-p']
    argv2 = ['', '-h']
    argv3 = ['', '-l']
    argv4 = ['', '-v']

    # Parse arguments
    Test = True
    arguments1 = parse_arguments(argv1, Test)
    arguments2 = parse_arguments(argv2, Test)
    arguments3 = parse_arguments(argv3, Test)
    arguments4 = parse_arguments(argv4, Test)

    print('User arguments:')
    print(arguments1)
    print(arguments2)
    print(arguments3)
    print(arguments4)


# ===========
# Script Main
# ===========

if __name__ == "__main__":
    test_parse_arguments()



# SPDX-FileCopyrightText: Copyright 2022, Siavash Ameli <sameli@berkeley.edu>
# SPDX-License-Identifier: BSD-3-Clause
# SPDX-FileType: SOURCE
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the license found in the LICENSE.txt file in the root directory
# of this source tree.

# =======
# Imports
# =======

import os
import sys
from datetime import date

# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

# The following is treatment exclusively for CYTHON. Since I build cython
# into '/build' directory, the lib files (*.so) are generated in the subfolder
# '/build/lib.linux-x86_64.3.8/'. To properly build sphinx, this directory
# should be included. Because this name of this subdirectory depends on the
# Linux platform, the architecture and the python version, in the following, we
# search for (glob) all subdirectories of '/build' and find which
# subdirectories contain '*.so' files. We then include all of those
# subdirectories to the path.

# Here as assumed that the '*.so' files are built inside the build directory.
# To do so,
# 1. Make sure cython package is built without '--inplace'. That is:
#    'python setup.py build_ext'.
# 2. Make sure in 'setup.cfg', the '[build_ext]' section does not have
#    'inplace=1' entry (if yes, comment it).

# If the build is make with '--inplace', then the '*.so' files are written
# inside the source code where '*.pyx' files are. In this case, you do not need
# to include the subdirectories of '/build' on the path.

# The RecursiveGolb.py should be located in '/docs'.
sys.path.insert(0, os.path.abspath('../'))
sys.path.insert(0, os.path.abspath('./'))
import recursive_glob  # this must be after including ./ path      # noqa: E402

# Build (assuming we build cython WITHOUT '--inplace', that is:
# 'python setup.py build_ext' only.
build_directory = os.path.join('..', 'build')

# Regex for pattern of lib files. Note: this is OS dependant. macos: *.dylib.
# Windows: *.dll
lib_file_patterns = ['*.so', '*.dylib', '*.dll']

# Find list of subdirectories of build directory that have files with pattern
build_subdirectories = recursive_glob.recursive_glob(
        build_directory,
        lib_file_patterns)

# Append the subdirectories to the path
for build_subdirectory in build_subdirectories:

    # Note: the subdirectory is *relative* to the BuildDirectory.
    path = os.path.join(build_directory, build_subdirectory)
    sys.path.insert(0, os.path.abspath(path))
    print(os.path.abspath(path))

# -- Project information -----------------------------------------------------

project = 'ortho'
author = 'Siavash Ameli'
copyright = f'{date.today().year}, ' + author

# -- Sphinx Settings ----------------------------------------------------------

# Check links and references
nitpicky = True

# Why setting "root_doc": by using toctree in index.rst, two things are added
# to index.html main page: (1) a toc in that location of page, (2) toc in the
# sidebar menu. If we add :hidden: option to toctree, it removes toc from
# both page and sidebar menu. There is no way we can have only one of these,
# for instance, toc only in the page, but not in the menu. A solution to
# this is as follows:
#   1. Set "root_doc= 'content'". Then add those toc that should go into the
#      menu in the content.rst file.
#   2. Add those toc that should go into the page in index.rst file.
# This way, we can control which toc appears where.
#
# A problem: by setting "root_doc='content'", the sidebar logo links to
# contents.html page, not the main page. There is a logo_url variable but it
# does not seem to do anything. To fix this, I added a javascript (see in
# /docs/source/_static/js/custom-pydata.js) which overwrites
# <a href"path/contents.html"> to <a href="path/index.html>".
# root_doc = "contents"

# Common definitions for the whole pages
rst_epilog = f'''
.. role:: synco
   :class: synco

.. |project| replace:: :synco:`ortho`
'''

# Figure, Tables, etc numbering
# numfig = True
# numfig_format = {
#     'figure': 'Figure %s',
#     'table': 'Table %s'
# }

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx_math_dollar',
    'sphinx.ext.mathjax',
    'sphinx.ext.graphviz', 'sphinx.ext.inheritance_diagram',
    # 'sphinx.ext.viewcode',
    'sphinx_toggleprompt',
    # 'sphinx.ext.autosectionlabel',
    'sphinx.ext.autosummary',
    # 'sphinx_automodapi.automodapi',
    # 'sphinxcontrib.napoleon',               # either use napoleon or numpydoc
    'numpydoc',                               # either use napoleon or numpydoc
    'sphinx_design',
    'sphinx-prompt',
    'sphinx_copybutton',
]

# Copy button settings
copybutton_prompt_is_regexp = True
copybutton_prompt_text = r'>>> |\.\.\. '

# Automatically generate autosummary after each build
autosummary_generate = True
autosummary_imported_members = True

# Remove the module names from the signature
# add_module_names = False

# automodapi
numpydoc_show_class_members = False

# Added after including sphinx_math_dollar. The following prevents msthjax to
# parse $ and $$.
mathjax3_config = {
    'tex2jax': {
        'inlineMath': [["\\(", "\\)"]],
        'displayMath': [["\\[", "\\]"]],
    },
}

# LaTeX
# 'sphinx.ext.imgmath',
# imgmath_image_format = 'svg'

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['build', 'Thumbs.db', '.DS_Store']


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
# Good themes
# html_theme = 'sphinx_rtd_theme'
html_theme = 'pydata_sphinx_theme'
# html_theme = 'sphinx_book_theme'
# html_theme = 'nature'
# html_theme = 'bizstyle'
# html_theme = 'haiku'
# html_theme = 'classic'

# Not good thmes
# html_theme = 'sphinxdoc'
# html_theme = 'alabaster'
# html_theme = 'pyramid'
# html_theme = 'agogo'
# html_theme = 'traditional'
# html_theme = 'scrolls'

# Options for theme
html_theme_options = {
    "github_url": "https://github.com/ameli/ortho",
    "navbar_end": [
        "theme-switcher",
        "search-field.html",
        "navbar-icon-links.html"
    ],
    "page_sidebar_items": ["page-toc", "edit-this-page"],
    # "header_links_before_dropdown": 4,
    "use_edit_page_button": True,
    "icon_links": [
        {
            "name": "PyPI",
            "url": "https://pypi.org/project/ortho/",
            "icon": "fab fa-python",
            "type": "fontawesome",
        },
        {
            "name": "Anaconda Cloud",
            "url": "https://anaconda.org/s-ameli/ortho",
            "icon": "fa fa-circle-notch",
            "type": "fontawesome",
        },
    ],
    "pygment_light_style": "tango",
    "pygment_dark_style": "native",
    "logo": {
        "image_light": "images/icons/logo-ortho-light.png",
        "image_dark": "images/icons/logo-ortho-dark.png",
    },
}

html_context = {
    "default_mode": "auto",
    "github_url": "https://github.com",
    "github_user": "ameli",
    "github_repo": "glearn",
    "github_version": "main",
    "doc_path": "docs/source",
}

html_sidebars = {
    "**": ["sidebar-nav-bs", "sidebar-ethical-ads"]
}

# Using Font Awesome icons
# html_css_files = [
#     "https://cdnjs.cloudflare.com/ajax/libs/font-awesome/4.7.0/css/font-awesome.min.css"
# ]


html_title = f"{project} Manual"
html_last_updated_fmt = '%b %d, %Y'
# html_show_sourcelink = False

# To use sphinx3, download at
# https://github.com/sphinx-doc/sphinx/tree/master/doc/_themes/sphinx13
# and put it in /docs/_themes
# html_theme = 'sphinx13'
# html_theme_path = ['_themes']

# import sphinx_readable_theme
# html_theme_path = [sphinx_readable_theme.get_html_theme_path()]
# html_theme = 'readable'

# import sphinx_nameko_theme
# html_theme_path = [sphinx_nameko_theme.get_html_theme_path()]
# html_theme = 'nameko'

# import sphinx_bootstrap_theme
# html_theme = 'bootstrap'
# html_theme_path = sphinx_bootstrap_theme.get_html_theme_path()

# extensions.append('sphinxjp.themes.basicstrap')
# html_theme = 'basicstrap'


# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# # Add css
# html_css_files = ["css/custom.css"]
# html_css_files = ['css/custom-anaconda-doc.css']

html_js_files = ["js/custom-pydata.css"]
# html_logo = '_static/images/icons/logo-imate-light.png'
html_favicon = '_static/images/icons/favicon.ico'


# =====
# setup
# =====

def setup(app):
    """
    This function is used to employ a css file to the themes.
    Note: paths are relative to /docs/source/_static
    """

    app.add_css_file('css/custom-pydata.css')
    app.add_js_file('js/custom-pydata.js')
    # app.add_css_file('css/custom.css')
    # app.add_css_file('css/custom-anaconda-doc.css')



#! /usr/bin/env python

# SPDX-FileCopyrightText: Copyright 2021, Siavash Ameli <sameli@berkeley.edu>
# SPDX-License-Identifier: BSD-3-Clause
# SPDX-FileType: SOURCE
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the license found in the LICENSE.txt file in the root directory
# of this source tree.


# =======
# Imports
# =======

# matplotlib without display
import matplotlib
matplotlib.use('Agg')

from ortho import OrthogonalFunctions                              # noqa: E402


# =========================
# Test Orthogonal Functions
# =========================

def test_orthogonal_functions():

    arguments = {
        'num_func': 9,
        'start_index': 1,
        'end_interval': 1
    }

    # Create an object
    OF = OrthogonalFunctions(**arguments)

    # Print the results
    OF.print()

    # Check the orthogonality of the functions
    OF.check(verbose=True)

    # Plot the results
    OF.plot()


# ===========
# Script Main
# ===========

if __name__ == "__main__":
    test_orthogonal_functions()



# SPDX-FileCopyrightText: Copyright 2021, Siavash Ameli <sameli@berkeley.edu>
# SPDX-License-Identifier: BSD-3-Clause
# SPDX-FileType: SOURCE
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the license found in the LICENSE.txt file in the root directory
# of this source tree.


# =======
# Imports
# =======

import sympy
import numpy
from .declarations import t


# ===
# phi
# ===

def phi(i):
    """
    Generates a list of non-orthogonal functions defined by
    :math:`t^{\\frac{1}{n}}`.

    :param i: The index of a basis function.
    :type i: int

    :return: A ``sympy`` object of the ``i``-th basis function.
    :rtype: sympy object
    """

    return t**(sympy.Rational(sympy.S(1), sympy.S(i+1)))


# =============
# Inner Product
# =============

def inner_product(f, g, interval):
    """
    Inner product of two functions with weight :math:`t^{-1}`.

    :param f: A ``sympy`` function.
    :type f: sympy object

    :param g: A ``sympy`` function.
    :type g: sympy object

    :param interval: The interval of the domain of the functions in the form
    ``[Start, end]``.
    :type interval: list

    :return: The inner product of the functions.
    :rtype: float
    """

    # The two stage sympy.expand below is needed so sympy.integrate can perform
    # properly
    h = sympy.expand(sympy.expand(f*g)/t)

    # Integrate function f between 0 and 1
    return sympy.integrate(h, (t, sympy.S(interval[0]), sympy.S(interval[1])))


# =========
# Normalize
# =========

def normalize(f, interval):
    """
    Normalize a function with respect to inner product.

    :param f: A sympy function.
    :type f: sympy object

    :param interval: The interval of the domain of the functions in the form
    ``[Start, end]``.
    :type interval: list

    :return: The normalized sympy function
    :rtype: sympy object
    """

    return f/sympy.sqrt(inner_product(f, f, interval))


# ====================
# Gram-Schmidt Process
# ====================

def gram_schmidt_process(
        num_func,
        start_index,
        interval,
        verbose=False):
    """
    Generates a list of orthonormalized symbolic functions.

    :param num_func: Number of functions to generate.
    :type num_func: int

    :param start_index: The start index of the functions.
    :type start_index: int

    :param interval: The interval of the domain of the functions in the form
        ``[Start, End]``.
    :type interval: list

    :return: list of sympy functions.
    :rtype: list
    """

    if verbose:
        print('---------------------')
        print('Orthogonal functions:')
        print('---------------------')
        print('')

    # Create an empty list of orthnormalized functions
    phi_orthonormalized_list = []

    # Gram-Schmidt orthogonalization process
    for i in range(num_func):

        # Initialize Gram-Schmidt process
        phi_non_orthogonal = phi(i+sympy.S(start_index))
        phi_orthogonalized = phi_non_orthogonal
        if i > 0:

            # Subtract a new function from each of previous orthonormalized
            # functions
            for j in range(i):

                # Subtract new non-orthogomal function from the projection of
                # the previous orthonormalized function
                phi_orthogonalized -= inner_product(
                        phi_non_orthogonal, phi_orthonormalized_list[j],
                        interval) * phi_orthonormalized_list[j]

        # Normalize an orthogonalized function
        phi_orthogonalized = sympy.simplify(phi_orthogonalized)
        phi_orthonormalized = normalize(phi_orthogonalized, interval)
        phi_orthonormalized = sympy.simplify(phi_orthonormalized)

        # Store result to the list
        phi_orthonormalized_list.append(phi_orthonormalized)

        # Print progress
        if verbose:
            print('phi_%d(t) = ' % (i+start_index))
            print(phi_orthonormalized_list[i])
            print('')

    return phi_orthonormalized_list


# ===========================
# Check Mutual Orthonormality
# ===========================

def check_mutual_orthonormality(
        phi_orthonormalized_list,
        interval,
        verbose=False):
    """
    Checks the inner orthonormality of each of two functions from a list of
    symbolic functions.

    It returns an array with elements ``-1``, ``0``, or ``1``.

        * ``-1``: two functions are not orthogonal.
        *  ``0``: two functions are orthogonal.
        * ``+1``: function is orthonormal.

    :param phi_orthonormalized_list: The list of sympy functions that are
        orthonormalized.
    :type phi_orthonormalized_list: list

    :param interval: The interval of the domain of the functions in the form
        ``[Start, End]``.
    :type interval: list

    :return: The mutual orthogonality matrix.
    :rtype: ndarray
    """

    # Initialize output
    num_func = len(phi_orthonormalized_list)
    mutual_inner_products = -1 * numpy.ones((num_func, num_func),
                                            dtype=int)

    # Mutual inner products
    for i in range(num_func):
        for j in range(i+1):

            # Inner product as symbolic number
            inner_prod = inner_product(phi_orthonormalized_list[i],
                                       phi_orthonormalized_list[j], interval)

            # Convert symbolic number to -1, 0, 1 numpy integer
            if inner_prod == sympy.S(1):
                mutual_inner_products[i, j] = 1
            elif inner_prod == sympy.S(0):
                mutual_inner_products[i, j] = 0

            # Symmetric matrix
            if i != j:
                mutual_inner_products[j, i] = mutual_inner_products[i, j]

    # Print results
    if verbose:
        print('----------------------------------')
        print('Mutual inner product of functions:')
        print('----------------------------------')
        print('')
        print(mutual_inner_products)
        print('')

    status = numpy.allclose(numpy.eye(num_func), mutual_inner_products)
    return status


# ===================
# get symbolic coeffs
# ===================

def get_symbolic_coeffs(
        phi_orthonormalized_list,
        start_index):
    """
    Gets symbolic coefficients of alpha[i] and a[ij].
    """

    num_func = len(phi_orthonormalized_list)
    sym_alpha = [None] * num_func
    sym_coeffs = [None] * num_func

    for j in range(num_func):

        # Multiply each function with sqrt(2/i+1) to have integer coeffs
        sym_alpha[j] = (sympy.S(-1)**(sympy.S(j))) * \
                 sympy.sqrt(sympy.Rational(2, j+start_index+1))
        function = phi_orthonormalized_list[j] / sym_alpha[j]
        function = sympy.simplify(function)

        # Convert the function to a polynomial
        polynomial = sympy.Poly(function)

        sym_coeffs[j] = []
        # Get the coeff of each monomial
        for i in range(j+1):
            coeff = polynomial.coeff_monomial(
                    t**(sympy.Rational(1, i+1+start_index)))
            sym_coeffs[j].append(coeff)

    return sym_alpha, sym_coeffs


# ==================
# Get numeric coeffs
# ==================

def get_numeric_coeffs(
        sym_alpha,
        sym_coeffs):
    """
    Evaluate symbolic coefficients to numerics.
    """

    num_func = len(sym_coeffs)
    alpha = []
    coeffs = [None] * num_func

    for j in range(num_func):
        alpha.append(float(sym_alpha[j]))

        coeffs[j] = []
        for i in range(len(sym_coeffs[j])):
            coeffs[j].append(int(sym_coeffs[j][i]))

    return alpha, coeffs


# =========================
# Print coeffs of functions
# =========================

def print_coeffs_of_functions(
        coeffs,
        start_index):
    """
    Prints the coeffs of orthonormalized functions as

    .. math::

        \\phi_j(t) = \\alpha_j * \\sum_{i=1}^n a_{ij} t^{\\frac{1}{i}}

    where :math:`\\alpha_j = \\sqrt{\\frac{2}{j}}`, and :math:`a_{ij}` are
    integers.

    :param phi_orthonormalized_list: The list of sympy functions that are
        orthonormalized.
    :type phi_orthonormalized_list: list

    :param start_index: The start index of the functions.
    :type start_index: int
    """

    print('-------------------------')
    print('coeff of functions:')
    print('-------------------------')
    print('')
    print('i       alpha_[i]     a_[ij]')
    print('------  -----------   ---------')

    num_func = len(coeffs)
    for j in range(num_func):

        # Print human friendly
        sign = (-1)**(j)
        sign_as_string = '-'
        if sign > 0:
            sign_as_string = '+'
        alpha_as_string = sign_as_string + \
            'sqrt(2/%d)' % (j+start_index+1)
        print('i = %d:  %11s  %s'
              % (j+start_index, alpha_as_string, coeffs[j]))

    print('')



#!/usr/bin/env python

# SPDX-FileCopyrightText: Copyright 2021, Siavash Ameli <sameli@berkeley.edu>
# SPDX-License-Identifier: BSD-3-Clause
# SPDX-FileType: SOURCE
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the license found in the LICENSE.txt file in the root
# directory of this source tree.

# =======
# Imports
# =======

from __future__ import print_function
import os
from os.path import join
import sys
import codecs
import subprocess


# ===============
# install package
# ===============

def install_package(package):
    """
    Installs packages using pip.

    Example:

    .. code-block:: python

        >>> install_package('numpy>1.11')

    :param package: Name of package with or without its version pin.
    :type package: string
    """

    subprocess.check_call([sys.executable, "-m", "pip", "install", package])


# =====================
# Import Setup Packages
# =====================

# Import setuptools
try:
    import setuptools
except ImportError:
    # Install setuptools
    install_package('setuptools')
    import setuptools


# =========
# read file
# =========

def read_file(Filename):
    with codecs.open(Filename, 'r', 'latin') as File:
        return File.read()


# ================
# read file to rst
# ================

def read_file_to_rst(Filename):
    try:
        import pypandoc
        rstname = "{}.{}".format(os.path.splitext(Filename)[0], 'rst')
        pypandoc.convert(read_file(Filename), 'rst', format='md',
                         outputfile=rstname)
        with open(rstname, 'r') as f:
            rststr = f.read()
        return rststr
    except ImportError:
        return read_file(Filename)


# ================
# get requirements
# ================

def get_requirements(directory, subdirectory=""):
    """
    Returns a list containing the package requirements given in a file named
    "requirements.txt" in a subdirectory.
    """

    requirements_filename = join(directory, subdirectory, "requirements.txt")
    requirements_file = open(requirements_filename, 'r')
    requirements = [i.strip() for i in requirements_file.readlines()]

    return requirements


# ====
# main
# ====

def main(argv):

    directory = os.path.dirname(os.path.realpath(__file__))
    package_name = "ortho"

    # Version
    version_dummy = {}
    version_file = join(directory, package_name, '__version__.py')
    exec(open(version_file, 'r').read(), version_dummy)
    version = version_dummy['__version__']
    del version_dummy

    # author
    author = open(os.path.join(directory, 'AUTHORS.txt'), 'r').read().rstrip()

    # Requirements
    requirements = get_requirements(directory)
    test_requirements = get_requirements(directory, subdirectory="tests")
    docs_requirements = get_requirements(directory, subdirectory="docs")

    # ReadMe
    long_description = open(os.path.join(directory, 'README.rst'), 'r').read()

    url = 'https://github.com/ameli/ortho'
    download_url = url + '/archive/main.zip'
    documentation_url = url + '/blob/main/README.rst'
    tracker_url = url + '/issues'

    # Setup
    setuptools.setup(
        name=package_name,
        version=version,
        author=author,
        author_email='sameli@berkeley.edu',
        description='Generate orthogonal set of functions',
        long_description=long_description,
        long_description_content_type='text/x-rst',
        keywords='orthogonal-functions regression sympy computer-algebra' +
                 'gram-schmidt',
        url=url,
        download_url=download_url,
        project_urls={
            "Documentation": documentation_url,
            "Source": url,
            "Tracker": tracker_url,
        },
        platforms=['Linux', 'OSX', 'Windows'],
        packages=setuptools.find_packages(exclude=[
            'tests.*',
            'tests',
            'examples.*',
            'examples']
        ),
        install_requires=requirements,
        python_requires='>=3.6',
        setup_requires=[
            'setuptools',
            'pytest-runner'],
        tests_require=[
            'pytest',
            'pytest-cov'],
        include_package_data=True,
        zip_safe=False,
        entry_points={
            "console_scripts": [
                "ortho=ortho.__main__:main"
            ]
        },
        extras_require={
            'test': test_requirements,
            'docs': docs_requirements,
        },
        classifiers=[
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9',
            'License :: OSI Approved :: BSD License',
            'Operating System :: OS Independent',
            'Natural Language :: English',
            'Intended Audience :: Science/Research',
            'Intended Audience :: Developers',
            'Topic :: Software Development',
            'Topic :: Software Development :: Libraries :: Python Modules',
        ],
    )


# ===========
# script main
# ===========

if __name__ == "__main__":
    main(sys.argv)



#! /usr/bin/env python

# SPDX-FileCopyrightText: Copyright 2021, Siavash Ameli <sameli@berkeley.edu>
# SPDX-License-Identifier: BSD-3-Clause
# SPDX-FileType: SOURCE
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the license found in the LICENSE.txt file in the root directory
# of this source tree.


# =======
# Imports
# =======

import sys

from OrthogonalFunctions._parse_arguments import parse_arguments
from OrthogonalFunctions import OrthogonalFunctions


# ====
# Main
# ====

def main(args=None):

    # User arguments
    if args is None:
        args = sys.argv

    # Parse arguments
    arguments = parse_arguments(args)

    # Create an object
    OF = OrthogonalFunctions(**arguments)

    # Compute the orthogonal functions
    OF.process()

    # Print the results
    OF.print()

    # Check the orthogonality of the functions
    if arguments['check_orthogonality']:
        OF.check()

    # Plot results
    if arguments['plot_flag']:
        OF.plot()


# ===========
# System Main
# ===========

if __name__ == "__main__":
    sys.exit(main())
